package com.restlet.client.async.impl;

import com.restlet.client.async.AsyncStreamWithResult;
import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.function.Consumer;
import com.restlet.client.utils.Nullable;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

public class WritableStreamWithResult<T, R> implements AsyncStreamWithResult<T, R> {

    private final List<T> stack = new ArrayList<>();
    private final List<Consumer<T>> consumers = new ArrayList<>();
    private final Deferred<R> closeDeferred;

    public WritableStreamWithResult() {
        this(Collections.<T>emptyList());
    }

    private WritableStreamWithResult(final List<T> initialValue) {
        stack.addAll(initialValue);
        closeDeferred = new DeferredImpl<R>();
    }

    public void accept(final List<T> list) throws IllegalStateException {
        for (final T val : list) {
            accept(val);
        }
    }

    public void accept(final T value) throws IllegalStateException {
        if (isClosed()) {
            throw new IllegalStateException("Stream is closed");
        }
        if (consumers.isEmpty()) {
            stack.add(value);
        } else {
            for (final Consumer<T> consumer : consumers) {
                consume(consumer, value);
            }
        }
    }

    @Override
    public AsyncStreamWithResult<T, R> forEach(final Consumer<T> consumer) {
        if (!isClosed()) {
            consumers.add(consumer);
        }
        if (stack.isEmpty()) {
            return this;
        }
        final Iterator<T> iterator = stack.iterator();
        while (iterator.hasNext()) {
            final T next = iterator.next();
            iterator.remove();
            consume(consumer, next);
        }
        return this;
    }

    private void consume(final Consumer<T> consumer, final T value) {
        try {
            consumer.consume(value);
        } catch (final Exception e) {
            //FIXME GWT.reportUncaughtException(new IllegalStateException("Consumer thrown an error when consuming value: " + value,e));
        }
    }

    @Override
    public Promise<List<T>> asList() {
        final List<T> returnValue = new ArrayList<>();
        if (isClosed()) {
            return Promises.of(returnValue);
        }
        forEach(new Consumer<T>() {
            @Override
            public void consume(final T t) {
                returnValue.add(t);
            }
        });

        return closeDeferred.promise().then(new PromiseHandler<Object>() {
            @Nullable
            @Override
            public Object on(@Nullable final Object param) {
                return returnValue;
            }
        });
    }

    @Override
    public AsyncStreamWithResult<T, R> closeIfStillOpened(final R result) {
        if (!isClosed()) {
            closeDeferred.resolve(result);
        }
        return this;
    }

    @Override
    public void closeIfStillOpened(final Throwable e) {
        if (!isClosed()) {
            closeDeferred.reject(e);
        }
    }

    @Override
    public void closeWhen(final Promise<R> closePromise) {
        closePromise
                .doOnResolve(new Consumer<R>() {
                    @Override
                    public void consume(final R result) {
                        closeIfStillOpened(result);
                    }
                })
                .doOnReject(new Consumer<Throwable>() {
                    @Override
                    public void consume(final Throwable throwable) {
                        closeIfStillOpened(throwable);
                    }
                });
    }

    @Override
    public boolean isClosed() {
        return closeDeferred.getState() != Deferred.State.PENDING;
    }

    @Override
    public Promise<R> closePromise() {
        return closeDeferred.promise();
    }

}
