package com.restlet.client.net.uri;

import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.Predicate;
import com.restlet.client.net.http.HttpClientException;
import com.restlet.client.net.http.request.impl.HttpRequest;
import com.restlet.client.net.uri.impl.UriSchemeToImpl;
import com.restlet.client.utils.Sequence;
import com.restlet.client.utils.StringUtils;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import static com.restlet.client.net.uri.UriUtils.SchemePrintOption.DONT_PRINT_SCHEME;
import static com.restlet.client.net.uri.UriUtils.SchemePrintOption.PRINT_SCHEME;
import static com.restlet.client.net.uri.UriUtils.SchemePrintOption.PRINT_SCHEME_DEFAULT;
import static com.restlet.client.net.uri.impl.UriSchemeToImpl.DEFAULT_SCHEME;
import static com.restlet.client.utils.Maybe.ofNullable;
import static com.restlet.client.utils.StringUtils.isBlank;
import static com.restlet.client.utils.StringUtils.trimLeft;

public class UriUtils {

    public static final String HTTP_SCHEME_AS_STRING = "http";
    public static final String HTTPS_SCHEME_AS_STRING = "https";
    public static final Character DEFAULT_QUERY_PARAMETER_SEPARATOR = '&';
    public static final Character DEFAULT_QUERY_PARAMETER_DELIMITER = '?';
    private static final String LOCAL_IPV6_ADDRESS = "::1";
    private static final Set<String> LOCAL_IPV4_ADDRESSES = new HashSet<>(Arrays.asList("localhost", "127.0.0.1", "0.0.0.0"));

    public static String replaceQuery(String url, final String query) {
        if (!isBlank(url)) {

            // strip fragment
            String fragment = "";
            final int findex = url.indexOf('#');
            if (findex != -1) {
                fragment = url.substring(findex);
                url = url.substring(0, findex);
            }

            final int qindex = url.indexOf(DEFAULT_QUERY_PARAMETER_DELIMITER);
            if (qindex != -1) {
                url = url.substring(0, qindex);
            }

            if (!isBlank(query) && !"?".equals(query)) {
                url = (query.startsWith("?"))
                        ? url.concat(query)
                        : url.concat("?").concat(query);
            }
            return url + fragment;
        }
        if (!isBlank(query)) {
            return query.startsWith("?") ? query : "?".concat(query);
        }
        return null;
    }

    public static String replaceAbsolutePath(String url, final String path) {
        if (!isBlank(url)) {
            final int index = url.indexOf('/');
            if (index == -1) {
                return !isBlank(path) ? url.concat(path) : url;
            }
            url = url.substring(0, index);
            return !isBlank(path) ? url.concat(path) : url;
        }
        return path;
    }

    public static String replaceRelativePath(String url, final String path) {
        if (!isBlank(url)) {
            final int index = url.lastIndexOf('/');
            if (index == -1) {
                return !isBlank(path) ? url.concat("/").concat(path) : url;
            }
            url = url.substring(0, index).concat("/");
            return !isBlank(path) ? url.concat(path) : url;
        }
        return path;
    }

    public static String printUri(final UriTo uri, final SchemePrintOption schemePrintOption) {
        if (uri == null) {
            return "";
        }

        final StringBuilder uriAsString = new StringBuilder();

        final String path = uri.getPath();
        if (schemePrintOption == PRINT_SCHEME && uri.getScheme() != null || schemePrintOption == PRINT_SCHEME_DEFAULT) {
            uriAsString.append(ofNullable(uri.getScheme()).orElse(DEFAULT_SCHEME).getName());
            if (isBlank(path) || !path.startsWith("://")) {
                uriAsString.append("://");
            }
        }

        uriAsString.append(ofNullable(uri.getHost()).orElse(""))
                .append(ofNullable(path).orElse(""));

        if (uri.getQuery() != null) {
            Sequence.of(uri.getQuery().getItems())
                    .filter(new Predicate<UriQueryParamTo>() {
                        @Override
                        public boolean test(final UriQueryParamTo value) {
                            return value != null && value.isEnabled() && !isBlank(value.getName());
                        }
                    })
                    .each(new BiConsumer<UriQueryParamTo, Integer>() {
                        @Override
                        public void consume(final UriQueryParamTo queryParameter, final Integer index) {
                            uriAsString.append(index == 0 ? "?" : uri.getQuery().getDelimiter());
                            if (!StringUtils.isBlank(queryParameter.getName())) {
                                uriAsString.append(queryParameter.getName());
                            }
                            if (!StringUtils.isBlank(queryParameter.getValue())) {
                                uriAsString.append("=")
                                        .append(queryParameter.getValue());
                            }
                        }
                    });
        }

        if (uri.getFragment() != null) {
            uriAsString.append("#").append(uri.getFragment());
        }
        return uriAsString.toString();
    }

    public static String extractHostPort(final String path) {
        if (path == null || path.trim().isEmpty()) {
            return path;
        }
        int i = path.indexOf('/');
        if (i == -1) {
            i = path.indexOf(DEFAULT_QUERY_PARAMETER_DELIMITER);
            if (i == -1) {
                i = path.indexOf('#');
            }
        }
        if (i != -1) {
            return path.substring(0, i).trim();
        }
        return path;
    }

    public static String addProtocolIfMissing(final String path) {
        String url = path;
        final String lowerCaseUrl = path.trim().toLowerCase();

        if (lowerCaseUrl.endsWith("/")) {
            url = url.substring(0, url.length() - 1);
        }
        if (!UriUtils.isHttp(lowerCaseUrl) && !UriUtils.isHttps(lowerCaseUrl)) {
            url = "http://" + url;
        }

        return url;
    }

    public static String getAbsolutePath(final UriTo uri) {
        final StringBuilder sb = new StringBuilder();

        if (uri.getScheme() == null) {
            sb.append(HTTP_SCHEME_AS_STRING);
        } else {
            sb.append(uri.getScheme().getName());
        }
        sb.append("://");
        if (!isBlank(uri.getHost())) {
            sb.append(uri.getHost());
        }
        if (!isBlank(uri.getPath())) {
            sb.append(uri.getPath());
        }
        return sb.toString();
    }

    public static boolean isHttp(final String uri) {
        return !isBlank(uri) && uri.trim().toLowerCase().startsWith("http://");
    }

    public static boolean isHttps(final String uri) {
        return !isBlank(uri) && uri.trim().toLowerCase().startsWith("https://");
    }

    public static boolean isHttpOrHttps(final String uri) {
        return isHttp(uri) || isHttps(uri);
    }

    public static boolean isMailto(final String uri) {
        return !StringUtils.isBlank(uri) && uri.trim().toLowerCase().startsWith("mailto:");
    }

    public static boolean isJavascript(final String uri) {
        return !StringUtils.isBlank(uri) && uri.trim().toLowerCase().startsWith("javascript:");
    }

    public static void ensureSchemeIsSet(final UriTo requestUri) {
        final String stringUri = printUri(requestUri, DONT_PRINT_SCHEME);

        if (isHttp(stringUri) || isHttps(stringUri)) {
            if (isHttps(requestUri.getHost())) {
                requestUri.setScheme(UriSchemeToImpl.valueOf("https"));
                requestUri.setHost(requestUri.getHost().substring("https://".length()));
                return;

            } else if (isHttp(requestUri.getHost())) {
                requestUri.setScheme(UriSchemeToImpl.valueOf("http"));
                requestUri.setHost(requestUri.getHost().substring("http://".length()));
                return;
            }

            if (isHttps(requestUri.getPath())) {
                requestUri.setScheme(UriSchemeToImpl.valueOf("https"));
                requestUri.setPath(requestUri.getPath().substring("https://".length()));

            } else if (isHttp(requestUri.getPath())) {
                requestUri.setScheme(UriSchemeToImpl.valueOf("http"));
                requestUri.setPath(requestUri.getPath().substring("http://".length()));
            }

        } else if (requestUri.getScheme() == null) {
            requestUri.setScheme(UriSchemeToImpl.valueOf(HTTP_SCHEME_AS_STRING));
        }
    }

    public static boolean isLoopbackAddress(final UriTo uri) {
        if (uri == null || isBlank(uri.getHost())) {
            return false;
        }

        final String uriString = UriUtils.printUri(uri, PRINT_SCHEME);

        try {
            final HttpRequest request = HttpRequest.parse(uriString.trim());
            if (request.isIp6()) {
                // Let's take care of IPv6 address, cf https://tools.ietf.org/html/rfc4291#section-2.2
                // We must remove leading 0 (aka compress the zeros in the RFC)
                final String[] groups = request.getPeerAddress().split(":");
                for (int i = 0; i < groups.length; i++) {
                    groups[i] = trimLeft(groups[i], '0');
                }
                return LOCAL_IPV6_ADDRESS.equals(StringUtils.joiner(groups).joinWith(":"));
            } else {
                return LOCAL_IPV4_ADDRESSES.contains(request.getPeerAddress());
            }
        } catch (final HttpClientException e) {
            return false;
        }
    }

    public enum SchemePrintOption {
        PRINT_SCHEME(true, true, false),
        PRINT_SCHEME_DEFAULT(true, true, true),
        DONT_PRINT_SCHEME(false, true, false),
        DONT_PRINT_HOST(false, false, false);

        public final boolean shouldPrintScheme;
        public final boolean shouldPrintHost;
        public final boolean shouldPrintDefaultSchemeIfNull;

        SchemePrintOption(final boolean shouldPrintScheme, final boolean shouldPrintHost,
                          final boolean shouldPrintDefaultSchemeIfNull) {
            this.shouldPrintScheme = shouldPrintScheme;
            this.shouldPrintHost = shouldPrintHost;
            this.shouldPrintDefaultSchemeIfNull = shouldPrintDefaultSchemeIfNull;
        }
    }
}
