package com.restlet.client.script.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.html.form.FormItemTo;
import com.restlet.client.html.form.FormItemType;
import com.restlet.client.model.EntityTo;
import com.restlet.client.net.http.request.HttpMethodDefinition;
import com.restlet.client.net.http.request.HttpRequestTo;
import com.restlet.client.net.http.request.impl.EffectiveHttpRequestTo;
import com.restlet.client.net.request.RequestHeaderTo;
import com.restlet.client.net.uri.UriQueryParamTo;
import com.restlet.client.net.uri.UriSchemeTo;
import com.restlet.client.net.uri.UriTo;
import com.restlet.client.net.uri.UriUtils;
import com.restlet.client.net.uri.impl.UriSchemeToImpl;
import com.restlet.client.utils.FunctionalUtils;
import com.restlet.client.utils.Maybe;
import com.restlet.client.utils.Mutator;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;
import com.restlet.client.utils.StringUtils;

import java.util.ArrayList;
import java.util.List;

import static com.restlet.client.async.Promises.of;
import static com.restlet.client.net.uri.impl.UriSchemeToImpl.isSupportedScheme;
import static com.restlet.client.utils.EntityCloner.ENTITY_CLONER;
import static com.restlet.client.utils.Objects.isNullOrEmpty;
import static com.restlet.client.utils.StringUtils.isBlank;

public abstract class RequestProcessor {

    private final Function<String, String> encoder = new Function<String, String>() {
        @Override
        public String apply(final String input) {
            return encodeQueryString(input);
        }
    };

    protected abstract Promise<String> process(String value);

    protected abstract String encodeQueryString(String query);

    public Promise<Boolean> isDynamic(final EntityTo entity) {
        if (entity instanceof HttpRequestTo) {
            // Here the clone is made to prevent changing the original request when processing it so that the InMemory value is not altered
            return process((HttpRequestTo) ENTITY_CLONER.cloneKeepParentsAsIs(entity))
                    .map(new Function<EffectiveHttpRequestTo, Boolean>() {
                        @Override
                        public Boolean apply(final EffectiveHttpRequestTo effectiveHttpRequestTo) {
                            return effectiveHttpRequestTo.isDynamic();
                        }
                    });
        }

        // FIXME dynamic does not make any sense for a container as stated by ARI
        return of(entity.getType() == EntityTo.Type.Scenario);
    }

    private boolean isLocal(final HttpRequestTo request) {
        return request != null && UriUtils.isLoopbackAddress(request.getUri());
    }

    @Mutator
    public Promise<EffectiveHttpRequestTo> process(final HttpRequestTo requestTo) {
        if (requestTo == null) {
            return of();
        }
        return evaluate(requestTo)
                .map(new Function<Integer, EffectiveHttpRequestTo>() {
                    @Override
                    public EffectiveHttpRequestTo apply(final Integer nbEvaluatedExpressions) {
                        return new EffectiveHttpRequestTo(requestTo, nbEvaluatedExpressions > 0, isLocal(requestTo));
                    }
                });
    }

    private Promise<Integer> evaluate(final HttpRequestTo effective) {
        final List<Promise<Integer>> nbDynamicExpressionsPromises = new ArrayList<>();

        // URI
        nbDynamicExpressionsPromises.add(evaluateUri(effective.getUri()));

        // method
        nbDynamicExpressionsPromises.add(evaluateMethod(effective.getMethod()));

        // headers
        nbDynamicExpressionsPromises.add(evaluateHeaders(effective.getHeaders()));

        // body
        nbDynamicExpressionsPromises.add(evaluateBody(effective));

        return Promises.all(nbDynamicExpressionsPromises)
                .map(new Function<List<Integer>, Integer>() {
                    @Override
                    public Integer apply(final List<Integer> input) {
                        return Sequence.of(input).reduce(new BiFunction<Integer, Integer, Integer>() {
                            @Override
                            public Integer apply(final Integer nbTotalDynamicExpressions, final Integer nbDynamicExpressions) {
                                return nbTotalDynamicExpressions + nbDynamicExpressions;
                            }
                        }, 0);
                    }
                });
    }

    private Promise<Integer> evaluateMethod(final HttpMethodDefinition method) {
        if (method == null || !method.isCustom()) {
            return of(0);
        }
        return process(method.getName()).map(new Function<String, Integer>() {
            @Override
            public Integer apply(final String evaluatedExpression) {
                final String originalMethodName = method.getName();
                method.setName(FunctionalUtils.ifNotNull(evaluatedExpression, new Function<String, String>() {
                    @Override
                    public String apply(final String input) {
                        return input.toUpperCase();
                    }
                }));
                return Objects.equals(originalMethodName, evaluatedExpression) ? 0 : 1;
            }
        });
    }

    private Promise<Integer> evaluateBody(final HttpRequestTo effective) {
        if (effective.getBody() == null
                || effective.getBody().getBodyType() == null
                || !effective.getMethod().isRequestBody()) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        switch (effective.getBody().getBodyType()) {
            case File:
                // do not evaluate a file
                break;
            case Form:
                if (effective.getBody().getFormBody() != null && !isNullOrEmpty(effective.getBody().getFormBody().getItems())) {
                    for (final FormItemTo item : effective.getBody().getFormBody().getItems()) {
                        if (item == null || !item.isEnabled()) {
                            continue;
                        }
                        if (!StringUtils.isBlank(item.getName())) {
                            dynamicExpressionsPromises.add(process(item.getName()).map(new Function<String, Boolean>() {
                                @Override
                                public Boolean apply(final String evaluatedExpression) {
                                    final String originalValue = item.getName();
                                    item.setName(evaluatedExpression);
                                    return !Objects.equals(originalValue, evaluatedExpression);
                                }
                            }));
                        }

                        if (FormItemType.Text.equals(item.getType()) && !StringUtils.isBlank(item.getValue())) {
                            dynamicExpressionsPromises.add(process(item.getValue()).map(new Function<String, Boolean>() {
                                @Override
                                public Boolean apply(final String evaluatedExpression) {
                                    final String originalValue = item.getValue();
                                    item.setValue(evaluatedExpression);
                                    return !Objects.equals(originalValue, evaluatedExpression);
                                }
                            }));
                        } else if (FormItemType.File.equals(item.getType())) {
                            // do not evaluate a file
                        }
                    }
                }
                break;

            case Text:
                if (!StringUtils.isBlank(effective.getBody().getTextBody())) {
                    dynamicExpressionsPromises.add(process(effective.getBody().getTextBody()).map(new Function<String, Boolean>() {
                        @Override
                        public Boolean apply(final String evaluatedExpression) {
                            final String originalValue = effective.getBody().getTextBody();
                            effective.getBody().setTextBody(evaluatedExpression);
                            return !Objects.equals(originalValue, evaluatedExpression);
                        }
                    }));
                }
                break;
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises);
    }

    private Promise<Integer> evaluateHeaders(final List<RequestHeaderTo> headers) {
        if (isNullOrEmpty(headers)) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        for (final RequestHeaderTo header : headers) {
            if (header == null || !header.isEnabled()) {
                continue;
            }

            if (!StringUtils.isBlank(header.getName())) {
                final String headerName = header.getName();
                dynamicExpressionsPromises.add(process(header.getName()).map(new Function<String, Boolean>() {
                    @Override
                    public Boolean apply(final String evaluatedExpression) {
                        header.setName(evaluatedExpression);
                        return !Objects.equals(headerName, evaluatedExpression);
                    }
                }));
            }
            if (!StringUtils.isBlank(header.getValue())) {
                final String headerValue = header.getValue();
                dynamicExpressionsPromises.add(process(header.getValue()).map(new Function<String, Boolean>() {
                    @Override
                    public Boolean apply(final String evaluatedExpression) {
                        header.setValue(evaluatedExpression);
                        return !Objects.equals(headerValue, evaluatedExpression);
                    }
                }));
            }
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises);
    }

    private Promise<Integer> evaluateUri(final UriTo requestUri) {
        if (requestUri == null) {
            return of(0);
        }

        final List<Promise<Boolean>> dynamicExpressionsPromises = new ArrayList<>();

        final UriSchemeTo originalScheme = requestUri.getScheme();

        if (originalScheme != null && !isSupportedScheme(originalScheme)) {
            dynamicExpressionsPromises.add(process(originalScheme.getName())
                    .map(new Function<String, Boolean>() {
                        @Override
                        public Boolean apply(final String evaluatedScheme) {
                            if (isSupportedScheme(evaluatedScheme)) {
                                requestUri.setScheme(UriSchemeToImpl.valueOf(evaluatedScheme));
                                return true;
                            } else {
                                requestUri.setScheme(null);
                                return !Objects.equals(evaluatedScheme, originalScheme.getName());
                            }
                        }
                    }));
        }

        // URI host
        if (!isBlank(requestUri.getHost())) {
            dynamicExpressionsPromises.add(process(requestUri.getHost()).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    final String originalValue = requestUri.getHost();
                    requestUri.setHost(evaluatedExpression);
                    return !Objects.equals(originalValue, evaluatedExpression);
                }
            }));
        }

        // URI path
        if (!isBlank(requestUri.getPath())) {
            dynamicExpressionsPromises.add(process(requestUri.getPath()).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    final String originalValue = requestUri.getPath();
                    requestUri.setPath(evaluatedExpression);
                    return !Objects.equals(originalValue, evaluatedExpression);
                }
            }));
        }

        // URI fragment
        if (!isBlank(requestUri.getFragment())) {
            dynamicExpressionsPromises.add(process(requestUri.getFragment()).map(new Function<String, Boolean>() {
                @Override
                public Boolean apply(final String evaluatedExpression) {
                    final String originalValue = requestUri.getFragment();
                    requestUri.setFragment(evaluatedExpression);
                    return !Objects.equals(originalValue, evaluatedExpression);
                }
            }));
        }

        // URI query
        if (requestUri.getQuery() != null && !isNullOrEmpty(requestUri.getQuery().getItems())) {

            for (final UriQueryParamTo qp : requestUri.getQuery().getItems()) {

                if (qp == null || !qp.isEnabled()) {
                    continue;
                }

                if (!isBlank(qp.getName())) {

                    dynamicExpressionsPromises.add(Promises.combine(process(qp.getName()), process(qp.getValue()), new Promises.CombineCallback<String, String, Boolean>() {
                        @Override
                        public Boolean on(final String name, final String value) {
                            final boolean isDynamic = !Objects.equals(qp.getName(), name) || !Objects.equals(qp.getValue(), value);

                            if (qp.isEncoded()) {
                                qp.setName(name);
                                qp.setValue(value);
                            } else {
                                qp.setName(Maybe.ofThrowable(encoder, name).orElse(name));
                                qp.setValue(Maybe.ofThrowable(encoder, value).orElse(value));
                            }

                            return isDynamic;
                        }
                    }));
                }

            }
        }

        return getNumberOfDynamicExpressions(dynamicExpressionsPromises)
                .doOnResolve(new Consumer<Integer>() {
                    @Override
                    public void consume(final Integer whatever) {
                        UriUtils.ensureSchemeIsSet(requestUri);
                    }
                });

    }

    private Promise<Integer> getNumberOfDynamicExpressions(final List<Promise<Boolean>> dynamicExpressionsPromises) {
        return Promises.all(dynamicExpressionsPromises)
                .map(new Function<List<Boolean>, Integer>() {
                    @Override
                    public Integer apply(final List<Boolean> dynamicExpressions) {
                        return Sequence.of(dynamicExpressions)
                                .filter(new Predicate<Boolean>() {
                                    @Override
                                    public boolean test(final Boolean isDynamic) {
                                        return isDynamic;
                                    }
                                })
                                .count();
                    }
                });
    }
}
