package com.restlet.client.script.runtime.impl;

import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.dao.EnvironmentsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.net.NamedValuedEntity;
import com.restlet.client.net.http.request.HttpMethodDefinition;
import com.restlet.client.net.request.RequestBodyTo;
import com.restlet.client.net.request.RequestHeaderTo;
import com.restlet.client.net.uri.UriQueryParamTo;
import com.restlet.client.net.uri.UriQueryTo;
import com.restlet.client.net.uri.UriTo;
import com.restlet.client.net.uri.UriUtils;
import com.restlet.client.platform.json.JsonEngine;
import com.restlet.client.platform.json.JsonException;
import com.restlet.client.platform.json.JsonObject;
import com.restlet.client.platform.json.JsonValue;
import com.restlet.client.script.ScriptEvaluator;
import com.restlet.client.utils.Sequence;
import com.restlet.client.utils.StringUtils;

import java.util.List;

import static com.restlet.client.net.http.HeaderUtils.toMessageHeaderTo;
import static com.restlet.client.net.uri.UriUtils.SchemePrintOption.PRINT_SCHEME;
import static com.restlet.client.script.runtime.impl.JsonReferenceUtils.buildJsonObjectForNamedValuedEntity;
import static com.restlet.client.utils.Objects.isNullOrEmpty;

public class RequestJSONReference {

    private final JsonObject root;
    private final JsonEngine jsonEngine;
    private final ScriptEvaluator scriptEvaluator;
    private final EnvironmentsDao environmentsDao;
    private final RepositoryDao repositoryDao;

    private RequestJSONReference(final JsonEngine jsonEngine,
                                 final ScriptEvaluator scriptEvaluator,
                                 final EnvironmentsDao environmentsDao,
                                 final RepositoryDao repositoryDao) {
        this.root = jsonEngine.newJsonObject();
        this.jsonEngine = jsonEngine;
        this.scriptEvaluator = scriptEvaluator;
        this.environmentsDao = environmentsDao;
        this.repositoryDao = repositoryDao;
    }

    public void setMethod(final HttpMethodDefinition method) {
        if (method != null) {
            root.put("method", jsonEngine.newJsonString(method.getName()));
        } else {
            root.put("method", null);
        }
    }

    public void setQuery(final UriQueryTo uriQuery) {
        if (uriQuery == null || isNullOrEmpty(uriQuery.getItems())) {
            root.put("query", jsonEngine.newJsonObject());
            return;
        }

        final JsonObject queryObject = buildJsonObjectForNamedValuedEntity(jsonEngine, Sequence.of(uriQuery.getItems())
                .map(new Function<UriQueryParamTo, NamedValuedEntity>() {
                    @Override
                    public NamedValuedEntity apply(final UriQueryParamTo queryParameter) {
                        return queryParameter;
                    }
                }).toList());
        root.put("query", queryObject);

    }

    public JsonValue asJSON() {
        return root;
    }

    public void setUri(final UriTo uri) {
        if (uri == null) {
            root.put("uri", null);
            return;
        }
        root.put("uri", jsonEngine.newJsonString(UriUtils.printUri(uri, PRINT_SCHEME)));

        setQuery(uri.getQuery());
    }

    public Promise<Void> setBody(final RequestBodyTo body) {
        if (body != null && body.getBodyType() != null) {
            switch (body.getBodyType()) {
                case File:
                case Form:
                    break;

                case Text:
                    if (!StringUtils.isBlank(body.getTextBody())) {
                        return scriptEvaluator.evaluate(body.getTextBody(), environmentsDao, repositoryDao)
                                .doOnResolve(new Consumer<String>() {
                                    @Override
                                    public void consume(final String parsedValue) {
                                        JsonValue json = null;
                                        try {
                                            json = RequestJSONReference.this.jsonEngine.fromJson(parsedValue);
                                        } catch (final JsonException ignored) {
                                        }

                                        if (json != null) {
                                            RequestJSONReference.this.root.put("body", json);
                                            return;
                                        }
                                        RequestJSONReference.this.root.put("body", RequestJSONReference.this.jsonEngine.newJsonString(body.getTextBody()));
                                    }
                                })
                                .castToVoidPromise();

                    }
                    break;
            }
        }
        return Promises.of();
    }

    public static Promise<RequestJSONReference> create(final JsonEngine jsonEngine,
                                                       final ScriptEvaluator scriptEvaluator,
                                                       final EnvironmentsDao environmentsDao,
                                                       final RepositoryDao repositoryDao,
                                                       final HttpMethodDefinition method,
                                                       final List<RequestHeaderTo> headers,
                                                       final UriTo uri,
                                                       final RequestBodyTo body) {
        final RequestJSONReference requestJSONReference = new RequestJSONReference(jsonEngine, scriptEvaluator, environmentsDao, repositoryDao);
        requestJSONReference.setMethod(method);
        requestJSONReference.root.put("headers", JsonReferenceUtils.getHeaders(jsonEngine, toMessageHeaderTo(headers)));
        requestJSONReference.setUri(uri);
        return requestJSONReference.setBody(body)
                .map(new Function<Void, RequestJSONReference>() {
                    @Override
                    public RequestJSONReference apply(final Void whatever) {
                        return requestJSONReference;
                    }
                });

    }
}
