package com.restlet.client.tests.impl;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.PromiseHandler;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.dao.EnvironmentsDao;
import com.restlet.client.dao.RepositoryDao;
import com.restlet.client.function.BiConsumer;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Executor;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.function.Supplier;
import com.restlet.client.log.LogService;
import com.restlet.client.log.LogService.Severity;
import com.restlet.client.model.EntityTo;
import com.restlet.client.model.EntityTreeNode;
import com.restlet.client.model.environment.EnvironmentTo;
import com.restlet.client.model.impl.RetryConfigurationImpl;
import com.restlet.client.net.http.HttpClient;
import com.restlet.client.net.http.HttpClientResultTo;
import com.restlet.client.net.http.HttpStatusCodeRange;
import com.restlet.client.net.http.request.HttpRequestTo;
import com.restlet.client.net.http.request.impl.EffectiveHttpRequestTo;
import com.restlet.client.net.http.response.HttpResponseTo;
import com.restlet.client.net.request.RetryConfiguration;
import com.restlet.client.script.impl.ExpressionRequestEvaluator;
import com.restlet.client.tests.ContainerTestResult;
import com.restlet.client.tests.SingleRequestTestResult;
import com.restlet.client.tests.SingleRequestTryTestResult;
import com.restlet.client.tests.TestResult;
import com.restlet.client.tests.TestResult.State;
import com.restlet.client.tests.asserts.AssertionEngine;
import com.restlet.client.tests.asserts.AssertionResultTo;
import com.restlet.client.tests.asserts.AssertionResultTo.AssertionState;
import com.restlet.client.tests.asserts.AssertionTo;
import com.restlet.client.tests.impl.ExecutionStateStore.RunType;
import com.restlet.client.utils.EntityUtils;
import com.restlet.client.utils.Holder;
import com.restlet.client.utils.Maybe;
import com.restlet.client.utils.Objects;
import com.restlet.client.utils.Sequence;
import com.restlet.client.utils.StringUtils;
import com.restlet.client.utils.TaskQueue;
import com.restlet.client.utils.TaskQueue.TaskConfiguration;
import com.restlet.client.utils.Timer.TimerRegistration;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import static com.restlet.client.async.Deferred.State.SUCCESS;
import static com.restlet.client.async.Promises.voidPromise;
import static com.restlet.client.model.EntityTo.Type.Request;
import static com.restlet.client.net.http.HttpStatusCodeRange.SUCCESS_CODE_RANGES;
import static com.restlet.client.net.http.HttpStatusCodeRange.ZERO;
import static com.restlet.client.tests.TestResult.State.Aborted;
import static com.restlet.client.tests.TestResult.State.Error;
import static com.restlet.client.tests.TestResult.State.Failure;
import static com.restlet.client.tests.TestResult.State.InProgress;
import static com.restlet.client.tests.TestResult.State.None;
import static com.restlet.client.tests.TestResult.State.Ok;
import static com.restlet.client.tests.TestResult.State.Skipped;
import static com.restlet.client.tests.impl.ExecutionStateStore.RunType.DB;
import static com.restlet.client.tests.impl.ExecutionStateStore.RunType.DRAFT;
import static com.restlet.client.utils.DateUtils.ONE_SECOND_IN_MS;
import static com.restlet.client.utils.HttpUtils.validateUriAndMethod;
import static com.restlet.client.utils.Objects.isNullOrEmpty;
import static com.restlet.client.utils.StringUtils.format;
import static com.restlet.client.utils.TaskQueue.TaskConfiguration.newTaskConfiguration;
import static com.restlet.client.utils.TaskQueue.newTaskQueue;
import static com.restlet.client.utils.Timer.TIMER;

public abstract class TestRunner {

    private static final List<AssertionState> ASSERTION_STATES_THAT_CONTAMINATE_REQUEST_STATE = Arrays.asList(AssertionState.Error, AssertionState.Failure, AssertionState.Ok);
    static final String MESSAGE_TEMPLATE_REQUEST_UNSUCCESSFUL_AFTER_N_TRIES = "The request did not succeed after %s tries.";

    private final TaskQueue taskQueue;
    private final HttpClient httpClient;
    private final ExpressionRequestEvaluator expressionRequestEvaluator;
    private final AssertionEngine assertionEngine;

    private final EnvironmentsDao environmentsDao;
    private final ExecutionStateStore executionStateStore;
    private final LogService logService;
    private final Severity severity;

    // Run-dependent
    private long startTime;
    private TimerRegistration elapsedTimeTimerRegistration;
    private TestResult currentTestResult;

    private boolean isAborting;
    private boolean isSkipping;
    private RunType runType;

    private DeferredImpl<Void> delayDeferred;
    private Promise<HttpClientResultTo> httpCallPromise;

    protected TestRunner(final HttpClient httpClient, final AssertionEngine assertionEngine,
                         final EnvironmentsDao environmentsDao,
                         final ExecutionStateStore executionStateStore, final LogService logService, final Severity severity,
                         final ExpressionRequestEvaluator expressionRequestEvaluator) {
        this.taskQueue = newTaskQueue("TEST RUNNER", logService, severity);
        this.httpClient = httpClient;
        this.assertionEngine = assertionEngine;
        this.environmentsDao = environmentsDao;
        this.executionStateStore = executionStateStore;
        this.logService = logService;
        this.severity = severity;
        this.expressionRequestEvaluator = expressionRequestEvaluator;
    }

    public Promise<Void> runRequestDontHonorRetry(final HttpRequestTo request, final RepositoryDao repositoryDao) {

        runType = DRAFT;
        executionStateStore.setEntityRunning(request);

        final SingleRequestTestResult requestResult = new SingleRequestTestResultImpl()
                .setEntity(request)
                .computeNameAndFullName();

        currentTestResult = requestResult;

        startAndNotifyParentRecursivelyIfRequired(requestResult, runType);
        executionStateStore.onRequestStarted(requestResult, runType);

        startElapsedTimeTimer(currentTestResult);

        final Holder<EffectiveHttpRequestTo> effectiveRequestHolder = Holder.empty();

        return expressionRequestEvaluator.effective(request, repositoryDao)
                .assignTo(effectiveRequestHolder)
                .map(new Function<EffectiveHttpRequestTo, EffectiveHttpRequestTo>() {
                    @Override
                    public EffectiveHttpRequestTo apply(final EffectiveHttpRequestTo effectiveRequest) {
                        requestResult.setDynamicRequest(effectiveRequest.isDynamic());
                        requestResult.setLocalRequest(effectiveRequest.isLocal());
                        requestResult.setEffectiveRequest(effectiveRequest);
                        return effectiveRequest;
                    }
                })
                .flatMap(new Function<EffectiveHttpRequestTo, Promise<Void>>() {
                    @Override
                    public Promise<Void> apply(final EffectiveHttpRequestTo effectiveRequest) {
                        validateRequestAndMarkAsInvalid(requestResult);
                        if (requestResult.isRequestInvalid()) {
                            return voidPromise();
                        }
                        return TestRunner.this.runSingleRequestOnceIfStateIsNotOk(1, 1, repositoryDao, effectiveRequest, requestResult, false);
                    }
                }).doFinally(new Consumer<Deferred<Void>>() {
                    @Override
                    public void consume(final Deferred<Void> value) {
                        executionStateStore.onRequestEnded(requestResult, runType);
                        executionStateStore.setEntityRunning(null);
                        elapsedTimeTimerRegistration.cancel();
                        taskQueue.empty();
                        isAborting = false;
                        isSkipping = false;
                    }
                });
    }

    public Promise<Void> run(final EntityTreeNode treeNode, final RepositoryDao repositoryDao, final EnvironmentTo environment) {
        if (treeNode == null) {
            throw new IllegalArgumentException("Cannot run null");
        }
        executionStateStore.setEntityRunning(treeNode.entity);
        runType = DB;

        treeNode.recursiveSortChildren();
        executionStateStore.recursiveClean(treeNode, repositoryDao);

        final boolean isRunningSingleRequest = treeNode.entity.getType() == Request;
        final List<Promise<Void>> promises = new ArrayList<>();
        final TestResult testResult;

        if (isRunningSingleRequest) {
            final HttpRequestTo request = (HttpRequestTo) treeNode.entity;
            final SingleRequestTestResult singleRequestTestResult = new SingleRequestTestResultImpl()
                    .setEntity(request)
                    .computeNameAndFullName();
            testResult = singleRequestTestResult;
            promises.add(taskQueue.addToQueue(planSingleRequestRun(singleRequestTestResult, repositoryDao)));

        } else {
            testResult = new ContainerTestResultImpl(executionStateStore)
                    .setState(None)
                    .setEntity(treeNode.entity)
                    .computeNameAndFullName();
            promises.addAll(Sequence.of(planContainerRun(treeNode, repositoryDao, environment))
                    .map(new Function<TaskConfiguration<Void>, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final TaskConfiguration<Void> taskConfiguration) {
                            return taskQueue.addToQueue(taskConfiguration);
                        }
                    })
                    .toList());
        }

        startElapsedTimeTimer(testResult);
        currentTestResult = testResult;
        executionStateStore.onElapsedTimeUpdated(testResult, runType);

        return Promises.all(promises)
                .castToVoidPromise()
                .doFinally(new Consumer<Deferred<Void>>() {
                    @Override
                    public void consume(final Deferred<Void> whatever) {
                        TestRunner.this.executionStateStore.setEntityRunning(null);
                        elapsedTimeTimerRegistration.cancel();
                        taskQueue.empty();
                        isAborting = false;
                        isSkipping = false;
                    }
                });
    }

    public void abort() {
        isAborting = true;
        elapsedTimeTimerRegistration.cancel();
        if (delayDeferred != null && delayDeferred.getState() != SUCCESS) {
            delayDeferred.resolve();
        }
        if (httpClient.abort()) {
            httpCallPromise.getDeferred().reject(new PromiseAbortedException());
        }
    }

    private void startElapsedTimeTimer(final TestResult testResult) {
        startTime = new Date().getTime();
        if (elapsedTimeTimerRegistration != null) {
            elapsedTimeTimerRegistration.cancel();
        }
        elapsedTimeTimerRegistration = TIMER.schedule(new Executor() {
            @Override
            public void execute() {
                testResult.setElapsedTime(new Date().getTime() - startTime);
                executionStateStore.onElapsedTimeUpdated(testResult, runType);
            }
        }, 1_000);
    }

    private List<TaskConfiguration<Void>> planContainerRun(final EntityTreeNode treeNode, final RepositoryDao repositoryDao, final EnvironmentTo environment) {
        final ArrayList<TaskConfiguration<Void>> taskConfigurations = new ArrayList<>();
        recursivePlanContainerRun(treeNode, repositoryDao, environment, null, taskConfigurations);
        return taskConfigurations;
    }

    private void recursivePlanContainerRun(final EntityTreeNode treeNode, final RepositoryDao repositoryDao,
                                           final EnvironmentTo environment, final ContainerTestResult parentResult,
                                           final ArrayList<TaskConfiguration<Void>> taskConfigurations) {
        final EntityTo entity = treeNode.entity;
        final TestResult testResult;
        switch (entity.getType()) {
            case Scenario:
                final ContainerTestResult scenarioResult = new ContainerTestResultImpl(executionStateStore)
                        .setEntity(entity)
                        .computeNameAndFullName()
                        .setParent(parentResult);

                testResult = scenarioResult;
                taskConfigurations.add(TestRunner.this.planScenarioRun(treeNode, scenarioResult, repositoryDao, environment));
                break;

            case Request:
                final HttpRequestTo request = (HttpRequestTo) entity;
                final SingleRequestTestResult requestResult = new SingleRequestTestResultImpl()
                        .setEntity(request)
                        .computeNameAndFullName()
                        .setParent(parentResult);

                testResult = requestResult;
                taskConfigurations.add(TestRunner.this.planSingleRequestRun(requestResult, repositoryDao));
                break;

            case Project:
            case Service:
            case LocalDriveRoot:
                final ContainerTestResult containerTestResult = new ContainerTestResultImpl(executionStateStore)
                        .setState(None)
                        .setEntity(entity)
                        .computeNameAndFullName()
                        .setParent(parentResult);

                testResult = containerTestResult;

                executionStateStore.index(containerTestResult, environment);

                //
                // `onContainerStarted` for this entity is either launched manually or by the children
                //
                if (isNullOrEmpty(treeNode.children)) {

                    taskConfigurations.add(newTaskConfiguration("Task for starting " + entity.getName(), new Supplier<Promise<Void>>() {
                        @Override
                        public Promise<Void> get() {
                            executionStateStore.onContainerStarted(containerTestResult, runType);
                            return Promises.voidPromise();
                        }
                    }));

                } else {

                    // children (scenarios or requests), launch `onContainerStarted` on their parents
                    Sequence.of(treeNode.children)
                            .each(new BiConsumer<EntityTreeNode, Integer>() {
                                @Override
                                public void consume(final EntityTreeNode childNode, final Integer index) {
                                    TestRunner.this.recursivePlanContainerRun(childNode, repositoryDao, environment, containerTestResult, taskConfigurations);
                                }
                            });

                }

                //
                // Once the children have added their task configurations, let's add an "end" task.
                //
                // It handles cases such as
                // - an empty project
                // - a project with empty children (services and scenarios)
                //
                // It is necessary only if the test result is not in a terminal state.
                //
                taskConfigurations.add(newTaskConfiguration("Task for ending " + entity.getName(), new Supplier<Promise<Void>>() {
                    @Override
                    public Promise<Void> get() {

                        if (!containerTestResult.getRunStatistics().result.isTerminal) {
                            containerTestResult.refreshState(runType, true);
                        }

                        return Promises.voidPromise();
                    }
                }));

                break;

            case Context:
            case RemoteUserDrive:
            case SharedDriveRoot:
                testResult = null;
                break;

            default:
                throw new IllegalArgumentException("Unlaunchable entity type: " + entity.getType());
        }

        if (parentResult != null && testResult != null) {
            parentResult.getTestResults().add(testResult);
        }

    }

    private TaskConfiguration<Void> planScenarioRun(final EntityTreeNode scenarioNode, final ContainerTestResult scenarioResult,
                                                    final RepositoryDao repositoryDao, final EnvironmentTo environment) {
        Sequence.of(scenarioNode.children)
                .each(new BiConsumer<EntityTreeNode, Integer>() {
                    @Override
                    public void consume(final EntityTreeNode requestNode, final Integer index) {
                        final HttpRequestTo request = (HttpRequestTo) requestNode.entity;
                        final SingleRequestTestResult requestResult = new SingleRequestTestResultImpl()
                                .setEntity(request)
                                .computeNameAndFullName()
                                .setParent(scenarioResult);
                        scenarioResult.getTestResults().add(requestResult);
                    }
                });
        executionStateStore.index(scenarioResult, environment);

        return newTaskConfiguration(format("Running scenario %s", scenarioResult.getFullName()), new Supplier<Promise<Void>>() {
            @Override
            public Promise<Void> get() {
                final TaskQueue scenarioTaskQueue = newTaskQueue(format("SCENARIO RUNNER %s", scenarioResult.getFullName()), logService, severity);

                startAndNotifyParentRecursivelyIfRequired(scenarioResult, runType);
                executionStateStore.onContainerStarted(scenarioResult, runType);

                final List<Promise<Void>> scenarioRequestsPromises = Sequence.of(scenarioResult.getTestResults())
                        .map(new Function<TestResult, Promise<Void>>() {
                            @Override
                            public Promise<Void> apply(final TestResult uncastRequestResult) {
                                final SingleRequestTestResult requestResult = (SingleRequestTestResult) uncastRequestResult;
                                return scenarioTaskQueue.addToQueue(format("Running requestResult %s", uncastRequestResult.getFullName()), new Supplier<Promise<Void>>() {
                                    @Override
                                    public Promise<Void> get() {
                                        return runSingleRequestHonorRetry(requestResult, repositoryDao);
                                    }
                                });
                            }
                        })
                        .toList();

                if (Objects.isNullOrEmpty(scenarioRequestsPromises)) {
                    scenarioResult.refreshState(runType, true);
                }

                return Promises.all(scenarioRequestsPromises)
                        .castToVoidPromise();

            }
        });
    }

    private void startAndNotifyParentRecursivelyIfRequired(final TestResult testResult, final RunType runType) {
        if (testResult.getParent() != null && testResult.getParent().getState() == None) {
            startAndNotifyParentRecursivelyIfRequired(testResult.getParent(), runType);
            testResult.getParent().setState(InProgress);
            executionStateStore.onContainerStarted(testResult.getParent(), runType);
        }
    }

    private TaskConfiguration<Void> planSingleRequestRun(final SingleRequestTestResult requestResult, final RepositoryDao repositoryDao) {
        return newTaskConfiguration(format("Running request %s", requestResult.getFullName()), new Supplier<Promise<Void>>() {
            @Override
            public Promise<Void> get() {
                return runSingleRequestHonorRetry(requestResult, repositoryDao);
            }
        });
    }

    private Promise<Void> runSingleRequestHonorRetry(final SingleRequestTestResult requestResult, final RepositoryDao repositoryDao) {
        executionStateStore.index(requestResult);

        startAndNotifyParentRecursivelyIfRequired(requestResult, runType);
        executionStateStore.onRequestStarted(requestResult, runType);

        final HttpRequestTo request = requestResult.getEntity();
        final RetryConfiguration retryConfiguration = Maybe.ofNullable(request.getRetryConfiguration())
                .orElse(new RetryConfigurationImpl());
        final Holder<EffectiveHttpRequestTo> effectiveRequestHolder = Holder.empty();

        final TaskQueue triesTaskQueue = newTaskQueue(format("TRIES OF REQUEST %s", requestResult.getFullName()), logService, severity);
        final List<Promise<Void>> triesPromises = new ArrayList<>();
        triesPromises.add(triesTaskQueue.addToQueue("Computing effective request", new Supplier<Promise<Void>>() {
            @Override
            public Promise<Void> get() {
                return (isAborting || shouldSkipRequest(requestResult) // No need to waste CPU
                        ? Promises.of(requestResult)
                        : expressionRequestEvaluator.effective(request, repositoryDao)
                        .assignTo(effectiveRequestHolder)
                        .map(new Function<EffectiveHttpRequestTo, EffectiveHttpRequestTo>() {
                            @Override
                            public EffectiveHttpRequestTo apply(final EffectiveHttpRequestTo effectiveRequest) {
                                requestResult.setDynamicRequest(effectiveRequest.isDynamic());
                                requestResult.setLocalRequest(effectiveRequest.isLocal());
                                requestResult.setEffectiveRequest(effectiveRequest);
                                validateRequestAndMarkAsInvalid(requestResult);
                                return effectiveRequest;
                            }
                        }))
                        .castToVoidPromise();
            }
        }));

        final int tryCount = retryConfiguration.isEnabled()
                ? retryConfiguration.getRetryCount() + 1
                : 1;

        for (int tryNumber = 1; tryNumber <= tryCount; tryNumber++) {
            final int genuineTryNumber = tryNumber;
            triesPromises.add(
                    triesTaskQueue.addToQueue(newTaskConfiguration(format("Try %s", tryNumber), new Supplier<Promise<Void>>() {
                        @Override
                        public Promise<Void> get() {
                            final Promise<Void> delayPromise;

                            final boolean hasRequestSucceeded = Sequence.of(requestResult.getTriesResults())
                                    .map(new Function<SingleRequestTryTestResult, Object>() {
                                        @Override
                                        public Object apply(final SingleRequestTryTestResult tryResult) {return tryResult.getState();}
                                    })
                                    .some(new Predicate<Object>() {
                                        @Override
                                        public boolean test(final Object state) {return state == Ok;}
                                    });
                            if (!shouldSkipRequest(requestResult) && !requestResult.isRequestInvalid() && !hasRequestSucceeded && !isAborting && genuineTryNumber > 1) {
                                delayDeferred = new DeferredImpl<>();
                                TIMER.scheduleOnce(new Executor() {
                                    @Override
                                    public void execute() {
                                        if (!delayDeferred.getState().isTerminal) {
                                            delayDeferred.resolve();
                                        }
                                    }
                                }, retryConfiguration.getIntervalInSeconds() * ONE_SECOND_IN_MS);
                                delayPromise = delayDeferred.promise();
                            } else {
                                delayPromise = voidPromise();
                            }

                            return delayPromise.flatMap(new Function<Void, Promise<Void>>() {
                                @Override
                                public Promise<Void> apply(final Void whatever) {
                                    return runSingleRequestOnceIfStateIsNotOk(genuineTryNumber, tryCount, repositoryDao, effectiveRequestHolder.value, requestResult, hasRequestSucceeded);
                                }
                            });

                        }
                    })));
        }

        return Promises.all(triesPromises)
                .castToVoidPromise()
                .doFinally(new Consumer<Deferred<Void>>() {
                    @Override
                    public void consume(final Deferred<Void> value) {
                        executionStateStore.onRequestEnded(requestResult, runType);
                        // event is propagated in order to send Mixpanel event
                        requestResult.refreshParentState(runType, true);
                    }
                });
    }

    private boolean shouldSkipRequest(final SingleRequestTestResult requestResult) {
        return requestResult.getParent() != null
                && requestResult.getParent().getState().isTerminal;
    }

    /**
     * One try to send the request, should only be called by {@link TestRunner#runSingleRequestHonorRetry(SingleRequestTestResult, RepositoryDao)}.
     * This method does not index the request in the execution state store !
     */
    private Promise<Void> runSingleRequestOnceIfStateIsNotOk(final int indexOfRetryOneBased, final int maxRetriesOneBased,
                                                             final RepositoryDao repositoryDao,
                                                             final EffectiveHttpRequestTo effectiveRequest,
                                                             final SingleRequestTestResult requestResult,
                                                             final boolean hasRequestSucceeded) {
        final boolean isLastTry = indexOfRetryOneBased == maxRetriesOneBased;
        requestResult.setCurrentTryBaseOneRetryIndex(indexOfRetryOneBased);

        final SingleRequestTryTestResult tryResult = new SingleRequestTryTestResultImpl()
                .setDisplayIndex(indexOfRetryOneBased)
                // Let's add the assertions by default so that the result at any given time is representative of the request
                .addAssertionResults(Sequence.of(requestResult.getEntity().getAssertions())
                        .map(new Function<AssertionTo, AssertionResultTo>() {
                            @Override
                            public AssertionResultTo apply(final AssertionTo assertion) {
                                return new AssertionResultTo(assertion);
                            }
                        })
                        .toList());

        requestResult.addTryResult(tryResult);

        Promise<Void> promise = voidPromise();
        if (isAborting) {
            if (isSkipping) {
                tryResult.setState(Skipped)
                        .updateParentIfRequestComplete(isLastTry);
            } else {
                isSkipping = true;
                tryResult.setState(Aborted)
                        .updateParentIfRequestComplete(isLastTry);
            }
        } else if (requestResult.isRequestInvalid()) {
            tryResult.setState(Skipped);
        } else if (shouldSkipRequest(requestResult)) {
            tryResult.setState(Skipped);
            requestResult.setState(Skipped);
        } else if (hasRequestSucceeded) {
            tryResult.setState(Skipped)
                    .updateParentIfRequestComplete(isLastTry);
        } else {
            executionStateStore.onRequestTryStarted(tryResult, runType);
            httpCallPromise = httpClient.send(effectiveRequest.getHttpRequestTo());
            promise = httpCallPromise
                    .flatMap(new Function<HttpClientResultTo, Promise<Void>>() {
                        @Override
                        public Promise<Void> apply(final HttpClientResultTo httpResult) {
                            if (httpResult == null) {
                                tryResult.setMessage("Unknown Error: null response")
                                        .setState(Error).updateParentIfRequestComplete(isLastTry);
                                return voidPromise();
                            }

                            tryResult.setHttpResult(httpResult);

                            if (TestRunner.this.hasAtLeastOneEnabledAssertion(effectiveRequest.getHttpRequestTo().getAssertions())) {
                                return TestRunner.this.evaluateAssertions(effectiveRequest.getHttpRequestTo(), httpResult, repositoryDao, tryResult)
                                        .doOnResolve(new Consumer<SingleRequestTryTestResult>() {
                                            @Override
                                            public void consume(final SingleRequestTryTestResult upToDateTryResult) {
                                                upToDateTryResult.updateParentIfRequestComplete(isLastTry);
                                            }
                                        })
                                        .castToVoidPromise();
                            }

                            tryResult.setState(Ok).updateParentIfRequestComplete(isLastTry);
                            final HttpResponseTo response = httpResult.getResponse();
                            if (response != null
                                    && response.getStatus() != null
                                    && response.getStatus().getCode() != null) {

                                if (ZERO.contains(response.getStatus().getCode())) {
                                    tryResult.setState(Error)
                                            .setMessage("The server may be unavailable")
                                            .updateParentIfRequestComplete(isLastTry);

                                } else if (!isValidResponseStatusCode(response)) {
                                    tryResult.setState(Failure)
                                            .setMessage("Response status code is " + response.getStatus().getCode())
                                            .updateParentIfRequestComplete(isLastTry);
                                }
                            }

                            return voidPromise();
                        }
                    })
                    .recoverOnReject(new PromiseHandler<Throwable>() {
                        @Override
                        public Object on(final Throwable error) {
                            if (error instanceof PromiseAbortedException) {
                                tryResult.setState(Aborted)
                                        .updateParentIfRequestComplete(isLastTry);
                                isSkipping = true;
                            } else {
                                tryResult.setState(Error)
                                        .setThrowable(error)
                                        .setMessage(error.getMessage())
                                        .updateParentIfRequestComplete(isLastTry);
                            }

                            return null;
                        }
                    });
        }

        return promise
                .flatMap(new Function<Void, Promise<Void>>() {
                    @Override
                    public Promise<Void> apply(final Void input) {
                        return TestRunner.this.executionStateStore.onRequestTryEnded(tryResult, runType);
                    }
                })
                .recoverOnReject();
    }

    private boolean isValidResponseStatusCode(final HttpResponseTo response) {
        return Sequence.of(SUCCESS_CODE_RANGES)
                .some(new Predicate<HttpStatusCodeRange>() {
                    @Override
                    public boolean test(final HttpStatusCodeRange acceptableStatusCode) {
                        return acceptableStatusCode.contains(response.getStatus().getCode());
                    }
                });
    }

    private boolean hasAtLeastOneEnabledAssertion(final List<AssertionTo> assertions) {
        return Sequence.of(assertions).some(new Predicate<AssertionTo>() {
            @Override
            public boolean test(final AssertionTo assertion) {
                return EntityUtils.isEnabled(assertion);
            }
        });
    }

    public Promise<Void> reevaluateAssertionsUponLastResponse(final HttpRequestTo request, final RepositoryDao repositoryDao) {
        final SingleRequestTestResult lastRequestResult = executionStateStore.getLastRequestResult(request.getId());
        if (lastRequestResult == null || lastRequestResult.getLastNonSkippedTry().isAbsent()) {
            return Promises.of();
        }

        final SingleRequestTryTestResult lastNonSkippedTry = lastRequestResult.getLastNonSkippedTry().get()
                .setState(InProgress);
        return evaluateAssertions(request, lastNonSkippedTry.getHttpResult(), repositoryDao, lastNonSkippedTry)
                .map(new Function<SingleRequestTryTestResult, Void>() {
                    @Override
                    public Void apply(final SingleRequestTryTestResult upToDateTryResult) {
                        upToDateTryResult.updateParentIfRequestComplete(true);
                        return null;
                    }
                });
    }

    private Promise<SingleRequestTryTestResult> evaluateAssertions(final HttpRequestTo effective, final HttpClientResultTo httpResult,
                                                                   final RepositoryDao repositoryDao, final SingleRequestTryTestResult tryResult) {

        final List<AssertionTo> latestAssertions = ((HttpRequestTo) repositoryDao.readSync(effective.getId())).getAssertions();
        final List<AssertionResultTo> assertionResults = Sequence.of(latestAssertions)
                .omitNulls()
                .map(new Function<AssertionTo, AssertionResultTo>() {
                    @Override
                    public AssertionResultTo apply(final AssertionTo assertion) {return new AssertionResultTo(assertion);}
                })
                .toList();

        if (isNullOrEmpty(assertionResults)) {
            if (isValidResponseStatusCode(tryResult.getHttpResult().getResponse())) {
                tryResult.setState(Ok);
            } else {
                tryResult.setState(Failure);
            }
        }

        tryResult.getAssertionsResults().clear();

        return assertionEngine.evaluateAndCheck(assertionResults, httpResult, environmentsDao, repositoryDao)
                .map(new Function<Void, SingleRequestTryTestResult>() {
                    @Override
                    public SingleRequestTryTestResult apply(final Void whatever) {
                        for (final AssertionResultTo assertionResult : assertionResults) {
                            if (!ASSERTION_STATES_THAT_CONTAMINATE_REQUEST_STATE.contains(assertionResult.getAssertionState())) {
                                continue;
                            }

                            if (tryResult.getState() == Failure
                                    && assertionResult.getAssertionState() == AssertionState.Error) {
                                // error status supersedes failure.
                                setStateAndMessageFromAssertion(tryResult, assertionResult);
                            } else if (tryResult.getState() == State.None
                                    || tryResult.getState() == InProgress
                                    || tryResult.getState() == Ok) {
                                setStateAndMessageFromAssertion(tryResult, assertionResult);
                            }
                        }
                        tryResult.addAssertionResults(assertionResults);
                        return tryResult;
                    }

                });
    }

    private void setStateAndMessageFromAssertion(final SingleRequestTryTestResult tryResult, final AssertionResultTo assertionResult) {
        switch (assertionResult.getAssertionState()) {
            case Error:
                completeResultMessage(tryResult, assertionResult, AssertionState.Error);
                tryResult.setState(Error);
                break;
            case Failure:
                completeResultMessage(tryResult, assertionResult, AssertionState.Failure);
                tryResult.setState(Failure);
                break;
            case Ok:
                tryResult.setState(Ok);
                break;
        }
    }

    private void completeResultMessage(final SingleRequestTryTestResult tryResult, final AssertionResultTo assertionResult,
                                       final AssertionState assertionState) {
        for (final AssertionResultTo.AssertionResultMessage message : assertionResult.getMessages(assertionState)) {
            tryResult.setMessage(message.message);
        }
    }

    private void validateRequestAndMarkAsInvalid(final SingleRequestTestResult requestResult) {
        final HttpRequestTo requestToSend = requestResult.getEffectiveRequest().getHttpRequestTo();
        final List<String> errors = validateUriAndMethod(requestToSend);
        if (!isNullOrEmpty(errors)) {
            requestResult.setRequestInvalid(true);
            requestResult.setState(Error);
            final SingleRequestTryTestResult tryResult = new SingleRequestTryTestResultImpl()
                    .setDisplayIndex(1);
            requestResult.addTryResult(tryResult);
            tryResult.setState(Error);
            tryResult.setMessage(StringUtils.joiner(errors).joinWith(" "));
            Sequence.of(requestResult.getEntity().getAssertions())
                    .omitNulls()
                    .each(new BiConsumer<AssertionTo, Integer>() {
                        @Override
                        public void consume(final AssertionTo assertion, final Integer index) {
                            tryResult.getAssertionsResults().add(new AssertionResultTo(assertion));
                        }
                    });
        }
    }
}
