package com.restlet.client.utils;

import com.restlet.client.function.BiFunction;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Function;
import com.restlet.client.function.Predicate;
import com.restlet.client.function.Supplier;

import java.io.Serializable;

public class Maybe<T> implements Serializable {

    final T value;

    protected Maybe(final T value) {
        this.value = value;
    }

    public Maybe() {
        this.value = null;
    }

    public <R> Maybe<R> get(final Function<T, R> getter) {
        return Maybe.of(
                isPresent()
                        ? getter.apply(get())
                        : null);
    }

    public static <R> Maybe<R> empty() {
        return new Maybe<>(null);
    }

    public static <R> Maybe<R> ofNullable(final R r) {
        if (r == null) {
            return empty();
        }
        return new Maybe<>(r);
    }

    public static <R> Maybe<R> ofThrowable(final Supplier<R> supplier) {
        try {
            return new Maybe<>(supplier.get());
        } catch (final Exception e) {
            return empty();
        }
    }

    public static <T, R> Maybe<R> ofThrowable(final Function<T, R> f, final T input) {
        try {
            return new Maybe<>(f.apply(input));
        } catch (final Exception e) {
            return empty();
        }
    }

    public static <T, U, R> Maybe<R> ofThrowable(final BiFunction<T, U, R> f, final T input1, final U input2) {
        try {
            return new Maybe<>(f.apply(input1, input2));
        } catch (final Exception e) {
            return empty();
        }
    }

    public static <R> Maybe<R> of(final R r) {
        return new Maybe<>(Objects.requireNonNull(r));
    }

    public <R> Maybe<R> map(final Function<T, R> mapper) {
        if (value == null) {
            return empty();
        }
        final R apply = mapper.apply(value);
        if (apply == null) {
            return empty();
        }
        return new Maybe<>(apply);
    }

    public <R> Maybe<R> flatMap(final Function<T, Maybe<R>> mapper) {
        if (value == null) {
            return empty();
        }
        final Maybe<R> apply = mapper.apply(value);
        if (apply == null) {
            return empty();
        }
        return apply;
    }

    /**
     * Returns the wrapped value. If value is null throw an exception
     *
     * @return unwrapped value
     */
    public T get() throws NullPointerException {
        Objects.requireNonNull(value);
        return value;
    }

    public T value() {
        return value;
    }

    public boolean isPresent() {
        return value != null;
    }

    public boolean isAbsent() {
        return value == null;
    }

    public T orElse(final T defaultValue) {
        if (value == null) {
            return defaultValue;
        }
        return value;
    }

    public Maybe<T> filter(final Predicate<T> predicate) {
        Objects.requireNonNull(predicate);
        if (isAbsent() || predicate.test(value)) {
            return this;
        } else {
            return empty();
        }
    }

    public Maybe<T> ifPresent(final Consumer<? super T> consumer) {
        if (value != null) {
            consumer.consume(value);
        }
        return this;
    }

    @Override
    public String toString() {
        return value == null ? "Nothing" : "Just " + String.valueOf(value);
    }
}
