package com.restlet.client.utils;

import com.restlet.client.async.Deferred;
import com.restlet.client.async.Promise;
import com.restlet.client.async.Promises;
import com.restlet.client.async.impl.DeferredImpl;
import com.restlet.client.function.Consumer;
import com.restlet.client.function.Executor;
import com.restlet.client.function.Predicate;
import com.restlet.client.function.Supplier;
import com.restlet.client.log.LogService;
import com.restlet.client.log.LogService.Severity;

import java.util.LinkedList;
import java.util.Queue;

import static com.restlet.client.async.Deferred.State.SUCCESS;
import static com.restlet.client.log.LogService.Severity.Debug;
import static com.restlet.client.utils.StringUtils.format;
import static com.restlet.client.utils.TaskQueue.TaskConfiguration.newTaskConfiguration;
import static com.restlet.client.utils.TaskQueue.Type.STANDARD;
import static com.restlet.client.utils.TaskQueue.Type.UNIQUE;

public class TaskQueue {

    private final String title;
    private final LogService logService;
    private final Severity severity;
    private final Queue<TaskConfiguration> queue = new LinkedList<>();

    public enum Type {
        STANDARD, UNIQUE
    }

    public static class TaskConfiguration<T> {
        final Type type;
        final String name;
        final Supplier<Promise<T>> taskPromiseSupplier;
        final Deferred<T> deferredResult = new DeferredImpl<>();
        Executor taskExecutor;

        private TaskConfiguration(final String name, final Type type, final Supplier<Promise<T>> taskPromiseSupplier) {
            this.type = type;
            this.name = name;
            this.taskPromiseSupplier = taskPromiseSupplier;
        }

        public static <T> TaskConfiguration<T> newTaskConfiguration(final String name, final Supplier<Promise<T>> taskPromiseSupplier) {
            return newTaskConfiguration(name, STANDARD, taskPromiseSupplier);
        }

        public static <T> TaskConfiguration<T> newTaskConfiguration(final String name, final Type type, final Supplier<Promise<T>> taskPromiseSupplier) {
            return new TaskConfiguration<>(name, type, taskPromiseSupplier);
        }
    }

    private TaskQueue(final String title, final LogService logService, final Severity severity) {
        this.title = title;
        this.logService = logService;
        this.severity = severity;
    }

    public static TaskQueue newTaskQueue(final String title, final LogService logService) {
        return newTaskQueue(title, logService, Debug);
    }

    public void empty() {
        final int removedTasksNumber = queue.size();
        queue.clear();
        logService.log(severity, format("Removed %s tasks from %s's queue", removedTasksNumber, title));
    }

    public static TaskQueue newTaskQueue(final String title, final LogService logService, final Severity severity) {
        return new TaskQueue(title, logService, severity);
    }

    public <T> Promise<T> addToQueue(final String taskTitle, final Supplier<Promise<T>> newPromiseSupplier) {
        return addToQueue(newTaskConfiguration(taskTitle, newPromiseSupplier));
    }

    public <T> Promise<T> addToQueue(final TaskConfiguration<T> taskConfiguration) {
        if (taskConfiguration.type == UNIQUE) {
            final boolean hasAlreadyOneUniqueTaskStacked = Sequence.of(queue)
                    .some(new Predicate<TaskConfiguration>() {
                        @Override
                        public boolean test(final TaskConfiguration stackedTask) {
                            return stackedTask.type == UNIQUE && Objects.equals(stackedTask.name, taskConfiguration.name);
                        }
                    });

            if (hasAlreadyOneUniqueTaskStacked) {
                return Promises.of();
            }
        }

        taskConfiguration.taskExecutor = getTaskExecutor(taskConfiguration);

        final boolean shouldRun = queue.isEmpty();

        queue.add(taskConfiguration);
        if (shouldRun) {
            taskConfiguration.taskExecutor.execute();
        }

        return taskConfiguration.deferredResult.promise();
    }

    private <T> Executor getTaskExecutor(final TaskQueue.TaskConfiguration<T> taskConfiguration) {
        return new Executor() {
            @Override
            public void execute() {
                logService.log(severity, format("%s BEGIN: %s", title, taskConfiguration.name));
                taskConfiguration.taskPromiseSupplier.get()
                        .doFinally(new Consumer<Deferred<T>>() {
                            @Override
                            public void consume(final Deferred<T> deferred) {
                                queue.poll();

                                if (deferred.getState() == SUCCESS) {
                                    logService.log(severity, format("%s END: %s", title, taskConfiguration.name));
                                    taskConfiguration.deferredResult.resolve((T) deferred.getResult());

                                } else {
                                    logService.error(format("%s ERROR: %s", title, taskConfiguration.name), deferred.getError());
                                    taskConfiguration.deferredResult.reject(deferred.getError());

                                }

                                if (!queue.isEmpty()) {
                                    queue.peek().taskExecutor.execute();
                                }
                            }
                        });
            }
        };
    }
}
